/*
Fast Gauss Bilateral Filtering C++ code 

 Shin Yoshizawa (shin@riken.jp)

Copyright:(c) Shin Yoshizawa, 2014.


 All right is reserved by Shin Yoshizawa.
This C++ code files are allowed for only primary user of research and educational purposes. Don't use secondary: copy, distribution, diversion, business purpose, and etc.. 

 In no event shall the author be liable to any party for direct, indirect, 
special, incidental, or consequential damage arising out of the use of this 
program and source files. 

 The C++ code and its simple manual are available at 
http://www.riken.jp/brict/Yoshizawa/Research/FBGFilter.html
*/

#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include"SimpleImage.h"
#include"fastgb.h"
#include"TifIO.h"

double getVariance(Image3D *in){
  double ave=0.0;
  double sdvi=0.0;
  int i,j,k;
  for(i=0;i<in->sz;i++)
    for(j=0;j<in->sy;j++)
      for(k=0;k<in->sx;k++){
	ave += in->img[i][j][k];
	sdvi += in->img[i][j][k]*in->img[i][j][k];
      
      }
  ave /= ((double)(in->sx*in->sy*in->sz));
  sdvi /= ((double)(in->sx*in->sy*in->sz));
  sdvi -= (ave*ave);
  
  return sqrt(sdvi);


}


//./Bilateral input.tif output.tif sigx sigy sigz sigI IRC error 
int main(int argc,char *argv[]){
 
  TifIO *mytifio = new TifIO(argv[1]);
  
if(mytifio->isApplicable()==false){
    printf("File \"%s\" is not supported.\n",mytifio->filename.c_str());
    return 1;
  }

 
 double sigX = atof(argv[3]);
  double sigY = atof(argv[4]);
  double sigZ = atof(argv[5]);
  double sigI = atof(argv[6]);
  
  int dr = atoi(argv[7]);
  double error = atof(argv[8]);

  if(sigX<=0.0||sigY<=0.0||sigZ<=0.0||sigI<=0.0||(!(dr==2||dr==3||dr==4||dr==5||dr==6))||error<=0.0){
    printf("Input parameter error !\n");
    printf("Error and sigmas should be greater than zero (>0).\n");
    printf("Also IRC should be 2, 3, 4, 5, or 6.\n");
    
    return 1;
  }

  
    if(mytifio->isColor()){
    Image3D *inR = new Image3D();
    Image3D *inG = new Image3D();
    Image3D *inB = new Image3D();
    mytifio->getColorImage(inR,inG,inB);
    
    
    
    double varR = getVariance(inR);
    double varG = getVariance(inG);
    double varB = getVariance(inB);
    
    printf("3D Color Image (%d,%d,%d): sigma(%lf,%lf,%lf) sdR = %lf sigIR %lf sdG = %lf sigIG %lf sdB = %lf sigIB %lf   IRC = %d error = %lf\n",inR->sx,inR->sy,inR->sz,sigX,sigY,sigZ,varR,(varR*sigI),varG,(varG*sigI),varB,(varB*sigI),dr,error);
    double sigIR = sigI*varR;
    double sigIG = sigI*varG;
    double sigIB = sigI*varB;
    
   double *sigma = new double[4];
  sigma[0] = sqrt(2.0*sigX*sigX);
  sigma[1] = sqrt(2.0*sigY*sigY);
  sigma[2] = sqrt(2.0*sigZ*sigZ);
  sigma[3] = sqrt(2.0*sigIR*sigIR);
  
  
  
  fastgb *myFGT = new fastgb(dr);
  
  myFGT->FastBilateralFilter(inR,sigma,error);
  delete myFGT;

  myFGT = new fastgb(dr);
  sigma[0] = sqrt(2.0*sigX*sigX);
  sigma[1] = sqrt(2.0*sigY*sigY);
  sigma[2] = sqrt(2.0*sigZ*sigZ);
  sigma[3] = sqrt(2.0*sigIG*sigIG);
  
  myFGT->FastBilateralFilter(inG,sigma,error);
  delete myFGT;
  
  myFGT = new fastgb(dr);
   sigma[0] = sqrt(2.0*sigX*sigX);
  sigma[1] = sqrt(2.0*sigY*sigY);
  sigma[2] = sqrt(2.0*sigZ*sigZ);
  sigma[3] = sqrt(2.0*sigIB*sigIB);
  
  myFGT->FastBilateralFilter(inB,sigma,error);
  delete myFGT;
  
    
  delete [] sigma;
  
        
    mytifio->setFilename(argv[2]);
    double maxR,minR;
    inR->setMaxMin(&maxR,&minR);
    double maxG,minG;
    inG->setMaxMin(&maxG,&minG);
    double maxB,minB;
    inB->setMaxMin(&maxB,&minB);
    if(maxR<maxB)maxR=maxB;
    if(maxR<maxG)maxR=maxG;
    
    mytifio->saveColorImage(inR,inG,inB,maxR,0.0,16);
    
    

    delete inR;
    delete inG;
    delete inB;
  }else{

  Image3D *in=NULL;
  
  
  
  
  
  
  in = new Image3D();
  
  mytifio->getImage(in);
  
  double var = getVariance(in);
  printf("3D Image (%d,%d,%d): sigma(%lf,%lf,%lf) sd = %lf sigI %lf  IRC = %d error = %lf\n",in->sx,in->sy,in->sz,sigX,sigY,sigZ,var,(var*sigI),dr,error);
  sigI *=var;
  double *sigma = new double[4];
  sigma[0] = sqrt(2.0*sigX*sigX);
  sigma[1] = sqrt(2.0*sigY*sigY);
  sigma[2] = sqrt(2.0*sigZ*sigZ);
  sigma[3] = sqrt(2.0*sigI*sigI);
  fastgb *myFGT = new fastgb(dr);
  
  myFGT->FastBilateralFilter(in,sigma,error);
  
  
  delete myFGT;
  
  delete [] sigma;
  
  
  mytifio->setFilename(argv[2]);
  double max,min;
  in->setMaxMin(&max,&min);
  mytifio->saveImage(in,max,0.0,16);
  
  
  
  
  if(in!=NULL)delete in;
    }
    delete mytifio;
  
  return 0;
}

