/*
Fast Gauss Bilateral Filtering C++ code 

 Shin Yoshizawa (shin@riken.jp)

Copyright:(c) Shin Yoshizawa, 2014.


 All right is reserved by Shin Yoshizawa.
This C++ code files are allowed for only primary user of research and educational purposes. Don't use secondary: copy, distribution, diversion, business purpose, and etc.. 

 In no event shall the author be liable to any party for direct, indirect, 
special, incidental, or consequential damage arising out of the use of this 
program and source files. 

 The C++ code and its simple manual are available at 
http://www.riken.jp/brict/Yoshizawa/Research/FBGFilter.html
*/

#define INF 1e100
//#define IRC 6
#define K1 1.089999
#define PI 3.141592653589793238462643383279502884197
#define NE 2.718281828459045235360287471352662497757
class gaussfgt4D{
 public:
  double *sgx;
  double *sgy;
  double *sgz;
  double *sgw;
  double sourcescale;
  double trxmax,trxmin,trymax,trymin,trzmax,trzmin,trwmax,trwmin;
  gaussfgt4D(){
    sgx=NULL;
    sgy=NULL;
    sgz=NULL;
    sgw=NULL;sourcescale=1.0;
    trxmax=trxmin=trymax=trymin=trzmax=trzmin=trwmax=trwmin=0.0;
  }
  virtual ~gaussfgt4D(){
    if(sgx!=NULL)free(sgx);
    if(sgy!=NULL)free(sgy);
    if(sgz!=NULL)free(sgz);
    if(sgw!=NULL)free(sgw);
  }
  

  inline double dmax(double a, double b) {
    return (a > b) ? a : b;
  }
  
  inline double dmin(double a, double b) {
    return (a < b) ? a : b;
  }
  
  inline double dabs(double d) {
    return (d > 0) ? d : -d;
  }
  
  inline double square(double d) {
    return d * d;
  }
  int getPWEK(double *weight,double eps,int natoms,double radii){
    // X. Wan & G. E.Karniadakis [JCP2006]
    // for 4D
    int i,ip,j;
    //double qsum=0.0;
    double val;
    double Kconst = K1*pow((2.0*PI),-0.25);
    //for(i=0;i<natoms;i++)qsum+=((double)(fabs(weight[i])));
    //if(qsum<=0.0)qsum=((double)(1.0));
    for(ip=1;;ip++){
      
      double rp=(radii*sqrt(NE/ip));
      double binomialC=0.0;
      double dval=Kconst*pow(ip,-0.25)*(pow(rp,ip)/(1.0-rp));
      val=0.0;
      for(j=1;j<=4;j++){
	if(j==1){
	  binomialC = 4.0;//(4!)/((4-j)!j!)
	}else if(j==2){
	  binomialC = 6.0;
	}else if(j==3){
	  binomialC = 4.0;
	}else if(j==4){
	  binomialC = 1.0;
	}
	val += binomialC*pow(dval,((double)(j)));
      }
      
      //val *= qsum;
      if(val<((double)(eps))){
	break;
      }
    }
    return ip;
  }
  // ******************************************************************
  void gausst(double* xat, double* yat, double* zat, double *wat,double* str, int
	      natoms,double* xtarg, double* ytarg, double* ztarg, double *wtarg,  double* pot, int ntarg,  double delta, double tol,int IRC) {
     
    int i, j, nside, nmax, kdis, nfmax, nlmax, /*nallbx,*/ ntbox;
    double /*toler,*/ boxdim, r=0, z;
    double qsum;
    int nterms=0;
    // dynamically allocated in lenchk
    int* iloc, *itofst, *itradr, *ibox;
    double* center, *****locexp;    
    double radii=0.5;
    
    boxdim = radii*sqrt(2.0*delta);
    if (natoms > ntarg) {
        nmax = natoms;
    } else {
        nmax = ntarg;
    }
    nside = (int) (1.0/boxdim)+1;
    nterms = getPWEK(str,tol,natoms,radii);
    kdis = IRC;
    
    nfmax = nterms + 2;
    nlmax = nterms + 2;
    lenchk(nside, xtarg, ytarg, ztarg, wtarg, ntarg, nterms, nmax, &iloc, &ibox, &itofst,
           &itradr, &locexp, &center, &ntbox);
    
    for(j=0;j<ntarg;j++)pot[j]=0.0;
    
    


    // call gafexp--
    gafexp(xat, yat, zat,wat, str, natoms, xtarg, ytarg, ztarg,wtarg, ntarg,
	   delta, nside, nterms, tol, iloc, ntbox, nfmax, nlmax, pot,
	   kdis, ibox, itradr, itofst, center, locexp);
    
     gaeval(xtarg, ytarg, ztarg,wtarg, pot, ntarg, locexp, nterms, ibox,
	   center, delta, iloc, nlmax, nside, itofst, itradr);
    
    
    
    
    
    freestuff(iloc, ibox, itofst, itradr, locexp, center, nterms, ntbox);
    return;
  }
    // ******************************************************************
  void gausst(double* xat, double* yat, double* zat, double *wat,double* str, int
	      natoms,  double* pot,  double delta, double tol,int IRC) {
     
    int i, j, nside, nmax, kdis, nfmax, nlmax, /*nallbx,*/ ntbox;
    double /*toler,*/ boxdim, r=0, z;
    double qsum;
    int nterms=0;
    // dynamically allocated in lenchk
    int* iloc, *itofst, *itradr, *ibox;
    double* center, *****locexp;    
    double radii=0.5;
    
    boxdim = radii*sqrt(2.0*delta);
    nmax = natoms;

    nside = (int) (1.0/boxdim)+1;
   
    nterms = getPWEK(str,tol,natoms,radii);
    
    kdis = IRC;
    //printf("nterms = %d, nside = %d, kdis = %d\n",nterms,nside,kdis);
    

    nfmax = nterms + 2;
    nlmax = nterms + 2;
    
    // allocate iloc, locexp, ibox, center, itofst, itradr and set up
    // iloc. the int ntbox is assigned there also
    lenchk(nside, xat, yat, zat,wat, natoms, nterms, nmax, &iloc, &ibox, &itofst,
           &itradr, &locexp, &center, &ntbox);
    
    
    
    
    for(j=0;j<natoms;j++)pot[j]=0.0;
    
    
    
    
    // call gafexp--
    
    gafexp(xat, yat, zat,wat, str, natoms, delta, nside, nterms, tol, iloc, ntbox, nfmax, nlmax, pot,
	   kdis, ibox, itradr, itofst, center, locexp);
    
    gaeval(xat, yat, zat,wat, pot, natoms, locexp, nterms, ibox,
	   center, delta, iloc, nlmax, nside, itofst, itradr);
    
    
    
    
    
    freestuff(iloc, ibox, itofst, itradr, locexp, center, nterms, ntbox);
    return;
}
 
// ******************************************************************
  void gdir(double xp, double yp, double zp,double wp, double* xat, double* yat,
	   double* zat,double* wat, int natoms, double* str, double* dpot, double
	   delta) {
    double x, y, z,w;
   int j;
   
   for (j=0;j<natoms;j++) {
     x = xp - xat[j];
     y = yp - yat[j];
     z = zp - zat[j];
     w = wp - wat[j];
     (*dpot) +=  exp(-(x*x + y*y + z*z +w*w)/delta)*str[j];
   }
  }
 
 // ******************************************************************
void gamval(double**** ffexp, int nterms, double* cent, double xpoint,
	    double ypoint, double zpoint,double wpoint, double* pot, double delta) {
  int i,j,k,l, i2;
  double facx, facy, facz, facw, x, y, z,w, x2, y2, z2,w2, dsq, potinc, hi, hj,hk;
  double* hexpx, *hexpy, *hexpz,*hexpw;
  
  hexpx = (double*) malloc((nterms+1) * sizeof(double));
  hexpy = (double*) malloc((nterms+1) * sizeof(double));
  hexpz = (double*) malloc((nterms+1) * sizeof(double));
  hexpw = (double*) malloc((nterms+1) * sizeof(double));
  
  dsq = 1.0/sqrt(delta);
  x = (xpoint - cent[0])*dsq;
  y = (ypoint - cent[1])*dsq;
  z = (zpoint - cent[2])*dsq;
  w = (wpoint - cent[3])*dsq;
  x2 = 2.0*x;
  y2 = 2.0*y;
  z2 = 2.0*z;
  w2 = 2.0*w;
  facx = exp(-x*x);
  facy = exp(-y*y);
  facz = exp(-z*z);
  facw = exp(-w*w);
  
  // inintialize
  hexpx[0] = facx;
  hexpy[0] = facy;
  hexpz[0] = facz;
  hexpw[0] = facw;
  
  hexpx[1] = x2*facx;
  hexpy[1] = y2*facy;
  hexpz[1] = z2*facz;
  hexpw[1] = w2*facw;
  
  for (i=1;i<nterms;i++) {
    i2 = 2*i;
    hexpx[i+1] = x2*hexpx[i] - i2*hexpx[i-1];
    hexpy[i+1] = y2*hexpy[i] - i2*hexpy[i-1];
    hexpz[i+1] = z2*hexpz[i] - i2*hexpz[i-1];
    hexpw[i+1] = w2*hexpw[i] - i2*hexpw[i-1];
    
  }
  
  potinc = 0.0;
  for (i=0;i<(nterms+1);i++) {
    hi = hexpx[i];
    for (j=0;j<(nterms+1);j++) {
      hj = hexpy[j];
      for (k=0;k<(nterms+1);k++) {
	hk = hexpz[k];
	for (l=0;l<(nterms+1);l++) {
	  potinc +=  (hi*hj*hk*hexpw[l]*ffexp[i][j][k][l]);
	}      
      }
    }
  }
  
  (*pot) +=  potinc;
  
  free(hexpx);
  free(hexpy);
  free(hexpz);
  free(hexpw);
  
}
 
// ******************************************************************
void gamkxp(double* cent, double* xatoms, double* yatoms, double* zatoms,double* watoms,
	    double* str, int ninbox, double**** ffexp, int nterms,
	    double delta) {
  int i,j,k,l,m;
  double dsq,x,y,z,w,xp;
  double* zp, *yp,*wp;
  wp = (double*) malloc((nterms+1) * sizeof(double));
  zp = (double*) malloc((nterms+1) * sizeof(double));
  yp = (double*) malloc((nterms+1) * sizeof(double));
  
  for (i=0;i<(nterms+1);i++) {
    for (j=0;j<(nterms+1);j++) {
      for (k=0;k<(nterms+1);k++) {
	for (l=0;l<(nterms+1);l++) {
	
	  ffexp[i][j][k][l] = 0.0;
	}
      }
    }
  }
  
  dsq = 1.0/sqrt(delta);
  for (i=0;i<ninbox;i++) {
    x = (xatoms[i] - cent[0])*dsq;
    y = (yatoms[i] - cent[1])*dsq;
    z = (zatoms[i] - cent[2])*dsq;
    w = (watoms[i] - cent[3])*dsq;
    wp[0] = 1.0;
    zp[0] = 1.0;
    yp[0] = 1.0;
    
    for (k=1;k<(nterms+1);k++) {
      wp[k] = wp[k-1]*w/k;
      zp[k] = zp[k-1]*z/k;
      yp[k] = yp[k-1]*y/k;
    }
    xp = str[i];
    for (j=0;j<(nterms+1);j++) {
      for (k=0;k<(nterms+1);k++) {
	for (l=0;l<(nterms+1);l++) {
	  for (m=0;m<(nterms+1);m++) {
	    ffexp[j][k][l][m] += (xp*yp[k]*zp[l]*wp[m]);
	  }
	}
      }
      xp *= (x/(j+1));
    } 
  }
  free(wp);
  free(zp);
  free(yp);
  
  return;
}

// ******************************************************************
 void galval(double**** local, int nterms, double* cent, double xpoint,
	     double ypoint, double zpoint, double wpoint, double* pot, double delta) {
   
   int i, j, k, l,m;
   double parsum1, parsum2, parsum3, parsum4, x, y, z,w, potinc, dsq;
   
   dsq = 1.0/(sqrt(delta));
   x = (xpoint - cent[0])*dsq;
   y = (ypoint - cent[1])*dsq;
   z = (zpoint - cent[2])*dsq;
   w = (wpoint - cent[3])*dsq;
   
   /* Debug 
   double ddx,ddy,ddz,ddw;
   potinc = 0.0;
   for(i=0;i<nterms+1;i++){
     ddx = pow(x,i);
     for(j=0;j<nterms+1;j++){
       ddy = pow(y,j);
       for(k=0;k<nterms+1;k++){
	 ddz = pow(z,k);
	 for(l=0;l<nterms+1;l++){
	   ddw = pow(w,l);
	   potinc= potinc + local[i][j][k][l]*ddx*ddy*ddz*ddw;
	 }
       }
     }
   }
   
   *pot = potinc;
   */
   /* using 4-d horner's rule */
   
   //loop x
   potinc = 0.0;
   parsum1=0.0;
   for(i=nterms;i>0;i--){
     
     //loop y
     parsum2=0.0;
     for(j=nterms;j>0;j--){
       
       //loop z
       parsum3=0.0;
       for(k=nterms;k>0;k--){
	 
	 //loop w
	 parsum4 = 0.0;
	 for(l=nterms;l>0;l--){
	   parsum4 = (parsum4+local[i][j][k][l])*w;
	 }
	 parsum4 += local[i][j][k][0];
	 parsum3 = (parsum3+parsum4)*z;
	 
       }
       parsum3 += local[i][j][0][0];
       parsum2 = (parsum2+parsum3)*y; 
     }
     
     //loop z
     parsum3=0.0;
     for(k=nterms;k>0;k--){
       //loop w
       parsum4 = 0.0;
       for(l=nterms;l>0;l--){
	 parsum4 = (parsum4+local[i][0][k][l])*w;
       }
       parsum4 += local[i][0][k][0];
       parsum3 = (parsum3+parsum4)*z;
     }
     
     
     //loop w
     parsum4 = 0.0;
     for(l=nterms;l>0;l--){
       parsum4 = (parsum4 + local[i][0][0][l])*w;
     }
     parsum4 += local[i][0][0][0];
     parsum1 = (parsum1 + parsum4 + parsum3+ parsum2)*x;
   }
   potinc += parsum1;
   
   
   
   
   
   //loop y
   parsum4 = 0.0;
   for(j=nterms;j>0;j--){
     //loop z
     parsum1 = 0.0;
     for(k=nterms;k>0;k--){
       //loop w
       parsum2 = 0.0;
       for(l=nterms;l>0;l--){
	 parsum2 = (parsum2 + local[0][j][k][l])*w;
       }
       parsum2 += local[0][j][k][0];
       parsum1 = (parsum1 + parsum2)*z;
     }
     
     //loop w
     parsum3 = 0.0;
     for(l=nterms;l>0;l--){
       parsum3 = (parsum3 + local[0][j][0][l])*w;
     }
     parsum3 += local[0][j][0][0];
     parsum4 = (parsum4 + parsum3 + parsum1)*y; 
     
   }
   potinc += parsum4;
   
   
   //loop z
   parsum1 = 0.0;
   for(k=nterms;k>0;k--){
     //loop w
     parsum2 = 0.0;
     for(l=nterms;l>0;l--){
       parsum2 = (parsum2 + local[0][0][k][l])*w;
     }
     parsum2 += local[0][0][k][0];
     parsum1 = (parsum1 + parsum2)*z;
   }
   potinc += parsum1;
   
   
   //loop w
   parsum1 = 0.0;
   for(l=nterms;l>0;l--){
     parsum1 = (parsum1 + local[0][0][0][l])*w;
   }
   parsum1 += local[0][0][0][0];
   potinc += parsum1;
   
   (*pot) = potinc;
   

   /* using 3-d horner's rule 
   potinc = 0.0;
   for(i=nterms;i>0;i--){
     parsum1 = 0.0;
     for(j=nterms;j>0;j--){
       parsum2 = 0.0;
       for(k=nterms;k>0;k--){
	 parsum2 = (parsum2 + local[i][j][k])*z;
       }
       parsum2 = parsum2 + local[i][j][0];
       parsum1 = (parsum1 + parsum2)*y;
     }
     parsum3 = 0.0;
     for(l=nterms;l>0;l--){
       parsum3 = (parsum3 + local[i][0][l])*z;
     }
     parsum3 = parsum3 + local[i][0][0];
     potinc = (potinc + parsum3 + parsum1)*x;
   }
   
   parsum1 = 0.0;
   for(j=nterms;j>0;j--){
     parsum2 = 0.0;
     for(k=nterms;k>0;k--){
       parsum2 = (parsum2 + local[0][j][k])*z;
     }
     parsum2 = parsum2 + local[0][j][0];
     parsum1 = (parsum1 + parsum2)*y;
   }
   potinc = potinc + parsum1;
   
   parsum1 = 0.0;
   for(k=nterms;k>0;k--){
     parsum1 = (parsum1 + local[0][0][k])*z;
    }
   parsum1 = parsum1 + local[0][0][0];
   potinc = potinc + parsum1;
   
   *pot = potinc;
   */
   /*

   //loop x
for(i=nterms;i>0;i--){
  
  //loop y
  for(j=nterms;j>0;j--){
    
    //loop z
    for(k=nterms;k>0;k--){
      
      //loop w
      for(l=nterms;l>0;l--){
      }
    }
  }
  
  //loop z
  for(k=nterms;k>0;k--){
    //loop w
    
    for(l=nterms;l>0;l--){
      
    }
    
  }
  
  
  //loop w
  
  for(l=nterms;l>0;l--){
    
  }
  
 }



//loop y
for(j=nterms;j>0;j--){
  
  //loop z
  for(k=nterms;k>0;k--){
    
    //loop w
    for(l=nterms;l>0;l--){
      
    }
  }
  //loop w
  
  for(l=nterms;l>0;l--){
    
  }
 }




//loop z
for(k=nterms;k>0;k--){
  //loop w
  for(l=nterms;l>0;l--){
  }
 }


//loop w
for(l=nterms;l>0;l--){
  
}



   */ 
    
    return;
 }
 

// ******************************************************************
void galkxp(double* cent, double xp, double yp, double zp, double wp, double
	    charge, double**** b, int nterms, double delta) {

  int i,l,m,n,o,i2;
    double dsq, x, y, z,w, x2, y2, z2,w2, facx, facy, facz,facw, /*prod,*/ facs, hl, hm,hn;

    double* hexpx, *hexpy, *hexpz, *hexpw;
 
    hexpx = (double*) malloc((nterms+1) * sizeof(double));
    hexpy = (double*) malloc((nterms+1) * sizeof(double));
    hexpz = (double*) malloc((nterms+1) * sizeof(double));
    hexpw = (double*) malloc((nterms+1) * sizeof(double));
    
    dsq = 1.0/sqrt(delta);

    // MAYBE WRONG!! I reversed them- makes sense see paper
    x = (cent[0] - xp)*dsq;
    x2 = 2.0*x;
    y = (cent[1] - yp)*dsq;
    y2 = 2.0*y;
    z = (cent[2] - zp)*dsq;
    z2 = 2.0*z;
    w = (cent[3] - wp)*dsq;
    w2 = 2.0*w;
    
    facx = exp(-x*x)*charge;
    facy = exp(-y*y);
    facz = exp(-z*z);
    facw = exp(-w*w);

    hexpx[0] = facx;
    hexpy[0] = facy;
    hexpz[0] = facz;
    hexpw[0] = facw;

    hexpx[1] = x2*facx;
    hexpy[1] = y2*facy;
    hexpz[1] = z2*facz;
    hexpw[1] = w2*facw;

    for (i=1;i<nterms;i++) {
        i2 = 2*i;
        hexpx[i+1] = x2*hexpx[i] - i2*hexpx[i-1];
        hexpy[i+1] = y2*hexpy[i] - i2*hexpy[i-1];
        hexpz[i+1] = z2*hexpz[i] - i2*hexpz[i-1];
	hexpw[i+1] = w2*hexpw[i] - i2*hexpw[i-1];
    }
  
    facs = 1.0;
    for (l=0;l<nterms+1;l++) {
        hexpx[l] = hexpx[l]*facs;
        hexpy[l] = hexpy[l]*facs;
        hexpz[l] = hexpz[l]*facs;
	hexpw[l] = hexpw[l]*facs;
        facs = -1*facs/(l+1.0);
    }
    for (l=0;l<nterms+1;l++) {
        hl = hexpx[l];
        for (m=0;m<nterms+1;m++) {
            hm = hexpy[m];
            for (n=0;n<nterms+1;n++) {
	      hn = hexpz[n];
              for(o=0;o<nterms+1;o++){
		b[l][m][n][o] = hl*hm*hn*hexpw[o];
	      }
            }
        }
    }
    free(hexpx);
    free(hexpy);
    free(hexpz);
    free(hexpw);
    
    return;
}

// ******************************************************************
void gaeval(double* xtarg, double* ytarg, double* ztarg, double* wtarg, double* pot,
	    int ntarg, double***** locexp, int nterms, int* ibox,
	    double* center, double delta, int* iloc, int nlmax, int
	    nside, int* itofst, int* itradr) {

    int i, k, jt, nboxes, inoff, jadr, ninbox;
    double xp, yp, zp,wp;

    nboxes = nside*nside*nside*nside;
    
    
    for (i=0;i<nboxes;i++) {
      inoff = itofst[i];
      ninbox = itofst[i+1] - inoff;
      if (ninbox > nlmax) {
	jadr = iloc[i];
	for (k=0; k< ninbox; k++) {
	  jt = itradr[inoff+k];
	  xp = xtarg[jt];
	  yp = ytarg[jt];
	  zp = ztarg[jt];
	  wp = wtarg[jt];
	  
	  galval(locexp[jadr], nterms, &center[(4*i)], xtarg[jt],
		 ytarg[jt], ztarg[jt], wtarg[jt], &pot[jt], delta);
	}
      }
    }
    return;
}



// ******************************************************************
 void gafexp(double* xat, double* yat, double* zat, double* wat, double* str, int natoms,double* xtarg,double* ytarg,double* ztarg,double* wtarg,int ntarg, double delta, int nside, int nterms, double tol,
	     int* iloc, int ntbox, int nfmax, int nlmax, double* pot,
	     int kdis, int* ibox, int* itradr, int* itofst, double*
	     center, double***** locexp)  {
    
  double* xatoms, *yatoms, *zatoms, *watoms, *charg2;
  double cent[4];
  double**** ffexp, ****b;
  double xp, yp, zp,wp;
  int i, j, k, l,m, nboxes, maxpts, ioff, iadr, inoff, nnbors, ninnbr, jt, iadloc, ninbox;
  int*ioffst, *nbors;
  double**** temp1, ****temp2, ****temp3;
  
  // temp1,temp2 not allocated in gshift to avoid overhead each time
  // it's called

  //  iloc, ibox, itradr, itofst, locexp, center allocated in
  //  lenchk. iloc is assigned there as well
  
  nboxes = nside*nside*nside*nside;
  if (natoms>ntarg) {
    maxpts = natoms;
  } else {
    maxpts = ntarg;
  }
    
    // dynamic allocations?
  ioffst = (int*) malloc((nboxes+1) * sizeof(int));
  nbors = (int*) malloc(nboxes * sizeof(int));
  xatoms = (double*) malloc(natoms * sizeof(double)); 
  yatoms = (double*) malloc(natoms * sizeof(double)); 
  zatoms  = (double*) malloc(natoms * sizeof(double)); 
  watoms  = (double*) malloc(natoms * sizeof(double)); 
  charg2  = (double*) malloc(natoms * sizeof(double)); 
  
  // locexp is allocated in lenchk
  
  // allocating the ffexp
  ffexp = (double****) malloc((nterms+1) * sizeof(double***));
  for (i=0;i<(nterms+1);i++) {
    ffexp[i] = (double ***) malloc((nterms+1) * sizeof(double**));
    for (j=0;j<(nterms+1);j++) {
      ffexp[i][j] = (double **) malloc((nterms+1) * sizeof(double *));
      for (k=0;k<(nterms+1);k++) {
	ffexp[i][j][k] = (double *) malloc((nterms+1) * sizeof(double));
	
      }
    }
  }
    
    
    // allocating b, temp1, temp2
    b = (double****) malloc((nterms+1) * sizeof(double***));
    temp1 = (double****) malloc((nterms+1) * sizeof(double***));
    temp2 = (double****) malloc((nterms+1) * sizeof(double***));
    temp3 = (double****) malloc((nterms+1) * sizeof(double***));
    for (i=0;i<(nterms+1);i++) {
      b[i] = (double***) malloc((nterms+1) * sizeof(double**));
        temp1[i] = (double***) malloc((nterms+1) * sizeof(double**));
        temp2[i] = (double***) malloc((nterms+1) * sizeof(double**));
	temp3[i] = (double***) malloc((nterms+1) * sizeof(double**));
        for (j=0;j<(nterms+1);j++) {
	  b[i][j] = (double**) malloc((nterms+1) * sizeof(double *));
	  temp1[i][j] = (double**) malloc((nterms+1) * sizeof(double *));
	  temp2[i][j] = (double**) malloc((nterms+1) * sizeof(double *));
	  temp3[i][j] = (double**) malloc((nterms+1) * sizeof(double *));
	  for (k=0;k<(nterms+1);k++) {
	    b[i][j][k] = (double*) malloc((nterms+1) * sizeof(double));
	    temp1[i][j][k] = (double*) malloc((nterms+1) * sizeof(double));
	    temp2[i][j][k] = (double*) malloc((nterms+1) * sizeof(double));
	    temp3[i][j][k] = (double*) malloc((nterms+1) * sizeof(double));
	    
	  }
        }
    }
    
    // call assign:
    assign(nside, xat, yat, zat,wat, str, xatoms, yatoms, zatoms, watoms, charg2,
           ioffst,xtarg, ytarg, ztarg,wtarg, ntarg,ibox, natoms, center,itradr, itofst); 
    
    

    // zero the local expansions
    for (i=0;i<ntbox;i++) {
      for (j=0;j<(nterms+1);j++) {
	for (k=0;k<(nterms+1);k++) {
	  for (l=0;l<(nterms+1);l++) {
	    for (m=0;m<(nterms+1);m++) {
	      locexp[i][j][k][l][m] = 0.0;
	    }
	  }
	}
      }
    }
    

    
    for (i=0;i<nboxes;i++) {
      ioff = ioffst[i];
      ninbox = ioffst[i+1] - ioff;
      if ((ninbox<=nfmax)&&(ninbox>0)) {
	// No far-field expansion
	mknbor(i, nbors, &nnbors, nside, kdis);
	for (j=0;j<nnbors;j++) {
	  iadr = nbors[j];
	  inoff = itofst[iadr];
	  ninnbr = itofst[iadr+1]-inoff;
	  if (ninnbr <= nlmax) {
	    // Direct interaction
	    for (k=0;k<ninnbr;k++) {
	      jt = itradr[inoff+k];
	      xp = xtarg[jt];
	      yp = ytarg[jt];
	      zp = ztarg[jt];
	      wp = wtarg[jt];
	      gdir(xp, yp, zp,wp, &xatoms[ioff], &yatoms[ioff],
		   &zatoms[ioff],&watoms[ioff], ninbox, &charg2[ioff],
		   &pot[jt], delta);
	    }
	  } else {
	    // Sources in neighbors shifted to taylor series
	    cent[0] = center[(4*iadr)];
	    cent[1] = center[(4*iadr+1)];
	    cent[2] = center[(4*iadr+2)];
	    cent[3] = center[(4*iadr+3)];
	    
	    iadloc = iloc[iadr];
	    for (k=0;k<ninbox;k++) {
	      galkxp(cent, xatoms[(ioff+k)], yatoms[(ioff+k)],
		     zatoms[(ioff+k)],watoms[(ioff+k)], charg2[(ioff+k)], b, nterms,
		     delta);
	      addexp(b, locexp[iadloc], nterms); 		    
	    }
	  }
	}
      } else if ((ninbox>nfmax)&&(ninbox>0)) {
            // create far-field expansion
	cent[0] = center[(4*i)];
	cent[1] = center[(4*i+1)];
	cent[2] = center[(4*i+2)];
	cent[3] = center[(4*i+3)];
            
	    gamkxp(cent, &xatoms[ioff], &yatoms[ioff], &zatoms[ioff], &watoms[ioff], &charg2[ioff], ninbox, ffexp, nterms, delta);
            mknbor(i, nbors, &nnbors, nside, kdis); 
            for(j=0;j<nnbors;j++){
	      iadr = nbors[j];
	      inoff = itofst[iadr];
	      ninnbr = itofst[iadr+1] - inoff;
	      if (ninnbr <= nlmax){ 
		// too few targets, shift to local
		for (k=0;k<ninnbr;k++) {
		  jt = itradr[(inoff+k)];
		  xp = xtarg[jt];
		  yp = ytarg[jt];
		  zp = ztarg[jt];
		  wp = wtarg[jt];
		  
		  gamval(ffexp,nterms,cent,xp,yp,zp,wp,&pot[jt],delta);
		}
	      }else{
		gshift(ffexp, cent, &center[(4*iadr)], b, nterms, temp1, temp2,temp3, delta); 
		iadloc = iloc[iadr];
		addexp(b, locexp[iadloc], nterms);
	      }
    	    }
      }
    }
    
    
    // free ffexp, b, temp1, temp2
    for(i = 0; i < (nterms+1); i++) {
      for (j=0; j < (nterms+1);j++) {
	for (k=0; k < (nterms+1);k++) {
	  free(ffexp[i][j][k]);
	  free(b[i][j][k]);
	  free(temp1[i][j][k]);
	  free(temp2[i][j][k]);
	  free(temp3[i][j][k]);
	}
	free(ffexp[i][j]);
	free(b[i][j]);
	free(temp1[i][j]);
	free(temp2[i][j]);
	free(temp3[i][j]);
      }
      free(ffexp[i]);
      free(b[i]);
      free(temp1[i]);
      free(temp2[i]);
      free(temp3[i]);
    }
    
    free(ffexp);
    free(b);
    free(temp1);
    free(temp2);
    free(temp3);

    free(xatoms);
    free(yatoms);
    free(zatoms);
    free(watoms);
    free(charg2);
    free(ioffst);
    free(nbors);
    return;
}

// ******************************************************************
 void gafexp(double* xat, double* yat, double* zat, double* wat, double* str, int
	     natoms, double delta, int nside, int nterms, double tol,
	     int* iloc, int ntbox, int nfmax, int nlmax, double* pot,
	     int kdis, int* ibox, int* itradr, int* itofst, double*
	     center, double***** locexp)  {
    
  double* xatoms, *yatoms, *zatoms, *watoms, *charg2;
  double cent[4];
  double**** ffexp, ****b;
  double xp, yp, zp,wp;
  int i, j, k, l,m, nboxes, maxpts, ioff, iadr, inoff, nnbors, ninnbr, jt, iadloc, ninbox;
  int*ioffst, *nbors;
  double**** temp1, ****temp2, ****temp3;
  
  // temp1,temp2 not allocated in gshift to avoid overhead each time
  // it's called

  //  iloc, ibox, itradr, itofst, locexp, center allocated in
  //  lenchk. iloc is assigned there as well
  
  nboxes = nside*nside*nside*nside;
  maxpts = natoms;
    
    // dynamic allocations?
  ioffst = (int*) malloc((nboxes+1) * sizeof(int));
  nbors = (int*) malloc(nboxes * sizeof(int));
  xatoms = (double*) malloc(natoms * sizeof(double)); 
  yatoms = (double*) malloc(natoms * sizeof(double)); 
  zatoms  = (double*) malloc(natoms * sizeof(double)); 
  watoms  = (double*) malloc(natoms * sizeof(double)); 
  charg2  = (double*) malloc(natoms * sizeof(double)); 
  
  // locexp is allocated in lenchk
  
  // allocating the ffexp
  ffexp = (double****) malloc((nterms+1) * sizeof(double***));
  for (i=0;i<(nterms+1);i++) {
    ffexp[i] = (double ***) malloc((nterms+1) * sizeof(double**));
    for (j=0;j<(nterms+1);j++) {
      ffexp[i][j] = (double **) malloc((nterms+1) * sizeof(double *));
      for (k=0;k<(nterms+1);k++) {
	ffexp[i][j][k] = (double *) malloc((nterms+1) * sizeof(double));
	
      }
    }
  }
    
    
    // allocating b, temp1, temp2
    b = (double****) malloc((nterms+1) * sizeof(double***));
    temp1 = (double****) malloc((nterms+1) * sizeof(double***));
    temp2 = (double****) malloc((nterms+1) * sizeof(double***));
    temp3 = (double****) malloc((nterms+1) * sizeof(double***));
    for (i=0;i<(nterms+1);i++) {
      b[i] = (double***) malloc((nterms+1) * sizeof(double**));
        temp1[i] = (double***) malloc((nterms+1) * sizeof(double**));
        temp2[i] = (double***) malloc((nterms+1) * sizeof(double**));
	temp3[i] = (double***) malloc((nterms+1) * sizeof(double**));
        for (j=0;j<(nterms+1);j++) {
	  b[i][j] = (double**) malloc((nterms+1) * sizeof(double *));
	  temp1[i][j] = (double**) malloc((nterms+1) * sizeof(double *));
	  temp2[i][j] = (double**) malloc((nterms+1) * sizeof(double *));
	  temp3[i][j] = (double**) malloc((nterms+1) * sizeof(double *));
	  for (k=0;k<(nterms+1);k++) {
	    b[i][j][k] = (double*) malloc((nterms+1) * sizeof(double));
	    temp1[i][j][k] = (double*) malloc((nterms+1) * sizeof(double));
	    temp2[i][j][k] = (double*) malloc((nterms+1) * sizeof(double));
	    temp3[i][j][k] = (double*) malloc((nterms+1) * sizeof(double));
	    
	  }
        }
    }
    
    // call assign:
    assign(nside, xat, yat, zat,wat, str, xatoms, yatoms, zatoms, watoms, charg2,
           ioffst,ibox, natoms, center,itradr, itofst); 
    
    

    // zero the local expansions
    for (i=0;i<ntbox;i++) {
      for (j=0;j<(nterms+1);j++) {
	for (k=0;k<(nterms+1);k++) {
	  for (l=0;l<(nterms+1);l++) {
	    for (m=0;m<(nterms+1);m++) {
	      locexp[i][j][k][l][m] = 0.0;
	    }
	  }
	}
      }
    }
    

    
    for (i=0;i<nboxes;i++) {
      ioff = ioffst[i];
      ninbox = ioffst[i+1] - ioff;
      if ((ninbox<=nfmax)&&(ninbox>0)) {
	// No far-field expansion
	mknbor(i, nbors, &nnbors, nside, kdis);
	for (j=0;j<nnbors;j++) {
	  iadr = nbors[j];
	  inoff = itofst[iadr];
	  ninnbr = itofst[iadr+1]-inoff;
	  if (ninnbr <= nlmax) {
	    // Direct interaction
	    for (k=0;k<ninnbr;k++) {
	      jt = itradr[inoff+k];
	      xp = xat[jt];
	      yp = yat[jt];
	      zp = zat[jt];
	      wp = wat[jt];
	      gdir(xp, yp, zp,wp, &xatoms[ioff], &yatoms[ioff],
		   &zatoms[ioff],&watoms[ioff], ninbox, &charg2[ioff],
		   &pot[jt], delta);
	    }
	  } else {
	    // Sources in neighbors shifted to taylor series
	    cent[0] = center[(4*iadr)];
	    cent[1] = center[(4*iadr+1)];
	    cent[2] = center[(4*iadr+2)];
	    cent[3] = center[(4*iadr+3)];
	    
	    iadloc = iloc[iadr];
	    for (k=0;k<ninbox;k++) {
	      galkxp(cent, xatoms[(ioff+k)], yatoms[(ioff+k)],
		     zatoms[(ioff+k)],watoms[(ioff+k)], charg2[(ioff+k)], b, nterms,
		     delta);
	      addexp(b, locexp[iadloc], nterms); 		    
	    }
	  }
	}
      } else if ((ninbox>nfmax)&&(ninbox>0)) {
            // create far-field expansion
	cent[0] = center[(4*i)];
	cent[1] = center[(4*i+1)];
	cent[2] = center[(4*i+2)];
	cent[3] = center[(4*i+3)];
            
	    gamkxp(cent, &xatoms[ioff], &yatoms[ioff], &zatoms[ioff], &watoms[ioff], &charg2[ioff], ninbox, ffexp, nterms, delta);
            mknbor(i, nbors, &nnbors, nside, kdis); 
            for(j=0;j<nnbors;j++){
	      iadr = nbors[j];
	      inoff = itofst[iadr];
	      ninnbr = itofst[iadr+1] - inoff;
	      if (ninnbr <= nlmax){ 
		// too few targets, shift to local
		for (k=0;k<ninnbr;k++) {
		  jt = itradr[(inoff+k)];
		  xp = xat[jt];
		  yp = yat[jt];
		  zp = zat[jt];
		  wp = wat[jt];
		  
		  gamval(ffexp,nterms,cent,xp,yp,zp,wp,&pot[jt],delta);
		}
	      }else{
		gshift(ffexp, cent, &center[(4*iadr)], b, nterms, temp1, temp2,temp3, delta); 
		iadloc = iloc[iadr];
		addexp(b, locexp[iadloc], nterms);
	      }
    	    }
      }
    }
    
    
    // free ffexp, b, temp1, temp2
    for(i = 0; i < (nterms+1); i++) {
      for (j=0; j < (nterms+1);j++) {
	for (k=0; k < (nterms+1);k++) {
	  free(ffexp[i][j][k]);
	  free(b[i][j][k]);
	  free(temp1[i][j][k]);
	  free(temp2[i][j][k]);
	  free(temp3[i][j][k]);
	}
	free(ffexp[i][j]);
	free(b[i][j]);
	free(temp1[i][j]);
	free(temp2[i][j]);
	free(temp3[i][j]);
      }
      free(ffexp[i]);
      free(b[i]);
      free(temp1[i]);
      free(temp2[i]);
      free(temp3[i]);
    }
    
    free(ffexp);
    free(b);
    free(temp1);
    free(temp2);
    free(temp3);

    free(xatoms);
    free(yatoms);
    free(zatoms);
    free(watoms);
    free(charg2);
    free(ioffst);
    free(nbors);
    return;
}

// ******************************************************************
void addexp(double****b, double****a, int nterms) {
  int i,j,k,l;
    for (i=0;i<nterms+1;i++) {
        for (j=0;j<nterms+1;j++) {
            for (k=0;k<nterms+1;k++) {
	      for (l=0;l<nterms+1;l++) {
	      
                a[i][j][k][l] +=  b[i][j][k][l];
	      }
            }
        }
    }
    return;
}


// ******************************************************************
void assign(int nside, double* xat, double* yat, double* zat, double* wat, double*
	    str, double* xatoms, double* yatoms, double* zatoms, double* watoms,
	    double* charg2, int* ioffst, double* xtarg, double*
	    ytarg, double* ztarg, double* wtarg, int ntarg, int* ibox, int natoms,
	    double* center, int* itradr, int* itofst) {
  // just to be clear, to avoid wierd 3d array effects, am leaving it
    // as a 3*nboxes length array.
  int j, nboxes, ixh, iyh, izh,iwh, iadr, indx;
  double h, hh;
  int* icnt; //allocate dynamically here
  int* icnt2; // same
  
  nboxes = nside*nside*nside*nside;
  icnt = (int*) malloc(nboxes * sizeof(int));
  icnt2 = (int*) malloc(nboxes * sizeof(int));
    
  for (j=0;j<nboxes;j++) {
    icnt2[j] = 0;
    icnt[j] = 0;
  }
  h = 1.0/nside;
   int nb2 =nside*nside;
   int nb3 = nb2*nside;
  
  for (j=0;j<natoms;j++) {
    ixh = (int) floor(xat[j]/h);
    if (ixh > nside - 1) {
      ixh = nside - 1;
    }
    if (ixh < 0) {
      ixh = 0;
    }
    iyh = (int) floor(yat[j]/h);
    if (iyh > nside -1) {
      iyh = nside - 1;
    }
    if (iyh < 0) {
      iyh = 0;
    }
    izh = (int) floor(zat[j]/h);
    if (izh > nside - 1) {
      izh = nside - 1;
    }
    if (izh < 0) {
      izh = 0;
    }
    iwh = (int) floor(wat[j]/h);
    if (iwh > nside - 1) {
      iwh = nside - 1;
    }
    if (iwh < 0) {
      iwh = 0;
    }
    
    iadr = iwh*nb3 +izh*nb2 + iyh*nside + ixh;
    icnt[iadr]++;
    ibox[j] = iadr;
  }
    
  ioffst[0] = 0;
  for (j=1;j<nboxes+1;j++) {
    ioffst[j] = ioffst[j-1] + icnt[j-1];
    
  }
  
  for (j=0;j<natoms;j++) {
    iadr = ibox[j];
    indx = ioffst[iadr] + icnt2[iadr];
    xatoms[indx] = xat[j];
    yatoms[indx] = yat[j];
    zatoms[indx] = zat[j];
    watoms[indx] = wat[j];
    charg2[indx] = str[j];
    //itradr[indx] = j;
    icnt2[iadr]++;
  }
  
  for (j=0;j<nboxes;j++) {
        icnt2[j] = 0;
        icnt[j] = 0;
    }
    for (j=0;j<ntarg;j++) {
        ixh = (int) floor(xtarg[j]/h);
        if (ixh > nside - 1) {
            ixh = nside - 1;
        }
        if (ixh < 0) {
            ixh = 0;
        }
        iyh = (int) floor(ytarg[j]/h);
        if (iyh > nside - 1) {
            iyh = nside - 1;
        }
        if (iyh < 0) {
            iyh = 0;
        }
        izh = (int) floor(ztarg[j]/h);
        if (izh > nside - 1) {
            izh = nside - 1;
        }
        if (izh < 0) {
            izh = 0;
        }

	iwh = (int) floor(wtarg[j]/h);
        if (iwh > nside - 1) {
            iwh = nside - 1;
        }
        if (iwh < 0) {
            iwh = 0;
        }
	
        iadr = iwh*nb3 +izh*nb2 + iyh*nside + ixh;
	icnt[iadr]++;
	ibox[j] = iadr;
    }

    itofst[0] = 0;
    for (j=1;j<nboxes+1;j++) {
      
      itofst[j] = itofst[j-1] + icnt[j-1];
    }
    for (j=0;j<ntarg;j++) {
      iadr = ibox[j];
      indx = itofst[iadr] + icnt2[iadr];
      itradr[indx] = j;
      icnt2[iadr]++;
    }
    
  hh = 0.5*h;
  for (j=0;j<4*nboxes;j=j+4) {
    
    iwh = (j/4)/(nb3);
    izh = ((j/4)%(nb3)) / nb2;
    iyh = (((j/4)%(nb3)) % nb2 )/ nside;
    ixh = (((j/4)%(nb3)) % nb2) % nside;
    
    center[j] = ixh*h + hh;
    center[j+1] = iyh*h + hh;
    center[j+2] = izh*h + hh;
    center[j+3] = iwh*h + hh;
  }
  free(icnt);
  free(icnt2);
  return;
}
  
// ******************************************************************
void assign(int nside, double* xat, double* yat, double* zat, double* wat, double*
	    str, double* xatoms, double* yatoms, double* zatoms, double* watoms,
	    double* charg2, int* ioffst, int* ibox, int natoms,
	    double* center, int* itradr, int* itofst) {
  // just to be clear, to avoid wierd 3d array effects, am leaving it
    // as a 3*nboxes length array.
  int j, nboxes, ixh, iyh, izh,iwh, iadr, indx;
  double h, hh;
  int* icnt; //allocate dynamically here
  int* icnt2; // same
  
  nboxes = nside*nside*nside*nside;
  icnt = (int*) malloc(nboxes * sizeof(int));
  icnt2 = (int*) malloc(nboxes * sizeof(int));
    
  for (j=0;j<nboxes;j++) {
    icnt2[j] = 0;
    icnt[j] = 0;
  }
  h = 1.0/nside;
   int nb2 =nside*nside;
   int nb3 = nb2*nside;
  
  for (j=0;j<natoms;j++) {
    ixh = (int) floor(xat[j]/h);
    if (ixh > nside - 1) {
      ixh = nside - 1;
    }
    if (ixh < 0) {
      ixh = 0;
    }
    iyh = (int) floor(yat[j]/h);
    if (iyh > nside -1) {
      iyh = nside - 1;
    }
    if (iyh < 0) {
      iyh = 0;
    }
    izh = (int) floor(zat[j]/h);
    if (izh > nside - 1) {
      izh = nside - 1;
    }
    if (izh < 0) {
      izh = 0;
    }
    iwh = (int) floor(wat[j]/h);
    if (iwh > nside - 1) {
      iwh = nside - 1;
    }
    if (iwh < 0) {
      iwh = 0;
    }
    
    iadr = iwh*nb3 +izh*nb2 + iyh*nside + ixh;
    icnt[iadr]++;
    ibox[j] = iadr;
  }
    
  itofst[0] = ioffst[0] = 0;
  for (j=1;j<nboxes+1;j++) {
    ioffst[j] = ioffst[j-1] + icnt[j-1];
    itofst[j] = itofst[j-1] + icnt[j-1];
  }
  
  for (j=0;j<natoms;j++) {
    iadr = ibox[j];
    indx = ioffst[iadr] + icnt2[iadr];
    xatoms[indx] = xat[j];
    yatoms[indx] = yat[j];
    zatoms[indx] = zat[j];
    watoms[indx] = wat[j];
    charg2[indx] = str[j];
    itradr[indx] = j;
    icnt2[iadr]++;
  }
  
  hh = 0.5*h;
  for (j=0;j<4*nboxes;j=j+4) {
    
    iwh = (j/4)/(nb3);
    izh = ((j/4)%(nb3)) / nb2;
    iyh = (((j/4)%(nb3)) % nb2 )/ nside;
    ixh = (((j/4)%(nb3)) % nb2) % nside;
    
    center[j] = ixh*h + hh;
    center[j+1] = iyh*h + hh;
    center[j+2] = izh*h + hh;
    center[j+3] = iwh*h + hh;
  }
  free(icnt);
  free(icnt2);
  return;
}
  


// ******************************************************************
void freestuff(int* iloc, int* ibox, int* itofst, int* itradr,
	       double***** locexp, double* center, int nterms, int ntbox) {
  
  // frees all the stuff allocated dynamically in lenchk
  int i,j,k,l;
  
  
  free(iloc);
  free(ibox);
  free(itofst);
  free(itradr);
  
  // free locexp
  for (i=0;i<ntbox;i++) {
    for (j=0;j<(nterms+1);j++) {
      for (k=0;k<(nterms+1);k++) {
	for (l=0;l<(nterms+1);l++) {
	  free(locexp[i][j][k][l]);
	}
	free(locexp[i][j][k]);
      }
      free(locexp[i][j]);
    }
    free(locexp[i]);
  }
  free(locexp);
  
  // center
  free(center);
  return;
}


// ******************************************************************
void gshift(double**** ffexp, double* cent1, double* cent2, double****
	    local, int nterms, double**** temp1, double**** temp2, double**** temp3, double delta) {

    int i,/*j,k,*/ len, i2, alpha1, alpha2, alpha3, alpha4, beta1, beta2, beta3, beta4;
    double dsq, x, y, z,w, x2, y2, z2,w2, sum, facx, facy, facz, facw;
    double* hexpx, *hexpy, *hexpz,*hexpw, *fac;

    hexpx = (double*) malloc((2*nterms+1) * sizeof(double));
    hexpy = (double*) malloc((2*nterms+1) * sizeof(double));
    hexpz = (double*) malloc((2*nterms+1) * sizeof(double));
    hexpw = (double*) malloc((2*nterms+1) * sizeof(double));
    
    fac = (double*) malloc((nterms+1) * sizeof(double));

    dsq = 1.0/sqrt(delta);
    x = (cent2[0]-cent1[0])*dsq;
    x2 = 2*x;
    y = (cent2[1]-cent1[1])*dsq;
    y2 = 2*y;
    z = (cent2[2]-cent1[2])*dsq;
    z2 = 2*z;
    w = (cent2[3]-cent1[3])*dsq;
    w2 = 2*w;

    facx = exp(-x*x);
    facy = exp(-y*y);
    facz = exp(-z*z);
    facw = exp(-w*w);

    hexpx[0] = facx;
    hexpy[0] = facy;
    hexpz[0] = facz;
    hexpw[0] = facw;

    hexpx[1] = x2*facx;
    hexpy[1] = y2*facy;
    hexpz[1] = z2*facz;
    hexpw[1] = w2*facw;

    for (i=1;i<(2*nterms);i++) {
        i2 = 2*i;
        hexpx[i+1] = x2*hexpx[i] - i2*hexpx[i-1];
        hexpy[i+1] = y2*hexpy[i] - i2*hexpy[i-1];
        hexpz[i+1] = z2*hexpz[i] - i2*hexpz[i-1];
	hexpw[i+1] = w2*hexpw[i] - i2*hexpw[i-1];
    }
    fac[0] = 1.0;
    for (i=1;i<nterms+1;i++) {
        fac[i] = -fac[i-1]/i;
    }

    len = nterms+1;


    
    for (alpha1=0;alpha1<len;alpha1++) {
      for (alpha2=0;alpha2<len;alpha2++) {
	for (alpha3=0;alpha3<len;alpha3++) {
	  for (beta4=0;beta4<len;beta4++) {
	    sum = 0.0;
	    for (alpha4=nterms;alpha4>=0;alpha4--) {
	      sum=sum+ffexp[alpha1][alpha2][alpha3][alpha4]*hexpw[alpha4+beta4];
	    }
	    temp1[alpha1][alpha2][alpha3][beta4] = sum;
	  }
        }
      }
    }

    
    for (alpha1=0;alpha1<len;alpha1++) {
      for (alpha2=0;alpha2<len;alpha2++) {
	for (beta3=0;beta3<len;beta3++) {
	  for (beta4=0;beta4<len;beta4++) {
	    sum = 0.0;
	    for (alpha3=nterms;alpha3>=0;alpha3--) {
	      sum=sum+temp1[alpha1][alpha2][alpha3][beta4]*hexpz[alpha3+beta3];
	    }
	    temp2[alpha1][alpha2][beta3][beta4] = sum;
	  }
	}
      }
    }
    
    for (alpha1=0;alpha1<len;alpha1++) {
      for (beta2=0;beta2<len;beta2++) {
	for (beta3=0;beta3<len;beta3++) {
	  for (beta4=0;beta4<len;beta4++) {
	    sum = 0.0;
	    for (alpha2=nterms;alpha2>=0;alpha2--) {
	      sum=sum+temp2[alpha1][alpha2][beta3][beta4]*hexpy[alpha2+beta2];
	    }
	    temp3[alpha1][beta2][beta3][beta4] = sum;
	  }
	}
      }
    }
    
    for (beta1=0;beta1<len;beta1++) {
      for (beta2=0;beta2<len;beta2++) {
	for (beta3=0;beta3<len;beta3++) {
	  for (beta4=0;beta4<len;beta4++) {
	    sum = 0.0;
	    for (alpha1=nterms;alpha1>=0;alpha1--) {
	      sum=sum+temp3[alpha1][beta2][beta3][beta4]*hexpx[alpha1+beta1];
	    }
	    local[beta1][beta2][beta3][beta4]=sum*fac[beta1]*fac[beta2]*fac[beta3]*fac[beta4];
	  }
        }
      }
    }
    
    /* Debug 
    for (beta1=0;beta1<len;beta1++) {
      for (beta2=0;beta2<len;beta2++) {
	for (beta3=0;beta3<len;beta3++) {
	  for (beta4=0;beta4<len;beta4++) {
	    sum = 0.0;
	    for (alpha1=0;alpha1<len;alpha1++) {
	      for (alpha2=0;alpha2<len;alpha2++) {
		for (alpha3=0;alpha3<len;alpha3++) {
		  for (alpha4=0;alpha4<len;alpha4++) {
		    sum = sum +ffexp[alpha1][alpha2][alpha3][alpha4]*hexpx[alpha1+beta1]*hexpy[alpha2+beta2]*hexpz[alpha3+beta3]*hexpw[alpha4+beta4]; 
		  }
		}
	      }
	    }
	    
	    local[beta1][beta2][beta3][beta4]=sum*fac[beta1]*fac[beta2]*fac[beta3]*fac[beta4];
	  }
	}
      }
    }
    */


    free(hexpx);
    free(hexpy);
    free(hexpz);
    free(hexpw);
    free(fac);
    return;
}


// ******************************************************************
void mknbor(int ibox, int* nbors, int* nnbors, int nside, double kdis) {

  int i, j, k,l, ihgt, irow, icol,izd, imin, imax, jmin, jmax, kmin, kmax,lmax,lmin;
  *nnbors = 0;
  int nd2=nside*nside;
  int nd3 = nside*nd2; 
  izd = ibox/nd3;
  ihgt = (ibox-izd*nd3)/nd2;
  //ihgt = ibox/(nside*nside);
  irow = ((ibox-izd*nd3) - ihgt*nd2)/nside;
  //irow = (ibox - ihgt*nside*nside)/nside; // same as modulo
  icol = ((ibox-izd*nd3) - ihgt*nd2)- irow*nside;
  //icol = (ibox - ihgt*nside*nside) - irow*nside; // ditto
  
  
  // threshold the x dimension
    if ((icol-kdis)>0)
        imin = (int) floor(icol-kdis);
    else
        imin = 0;
    if ((icol+kdis)<(nside-1))
        imax = (int) floor(icol + kdis);
    else
        imax = nside-1;
    // threshold y
    if ((irow-kdis)>0)
        jmin = (int) floor(irow - kdis);
    else
        jmin = 0;
    if ((irow+kdis)<(nside-1))
        jmax = (int) floor(irow+kdis);
    else 
        jmax = nside-1;
    // threshold z
    if ((ihgt-kdis)>0)
        kmin = (int) floor(ihgt-kdis);
    else
        kmin = 0;
    if ((ihgt+kdis)<(nside-1))
        kmax = (int) floor(ihgt+kdis);
    else
        kmax = nside-1;
    if ((izd-kdis)>0)
        lmin = (int) floor(izd-kdis);
    else
        lmin = 0;
    if ((izd+kdis)<(nside-1))
        lmax = (int) floor(izd+kdis);
    else
        lmax = nside-1;
    
    
    for (i=imin;i<imax+1;i++) {
      for (j=jmin;j<jmax+1;j++) {
	for (k=kmin;k<kmax+1;k++) {
	  for (l=lmin;l<lmax+1;l++) {
	    nbors[*nnbors] = l*nd3+k*nd2+j*nside+i;
	    (*nnbors)++;
	  }
	}
      }
    }
    return;
}


// ******************************************************************
void lenchk(int nside, double* xtarg, double* ytarg, double* ztarg,double* wtarg,
	    int ntarg, int nterms, int nmax, int** iloc, int** ibox,
	    int** itofst, int** itradr, double****** locexp, double**
	    center, int* ntbox) {

  int i, j, k,l, nboxes, ixh, iyh, izh,iwh, iadr;
  double h;
  
  nboxes = nside*nside*nside*nside;
  (*ntbox) = 0;
  //*ntbox = 0;
  // allocate the arrays
  *iloc = (int*) malloc(nboxes * sizeof(int));
  *itofst = (int*) malloc((nboxes+1) * sizeof(int));
  *itradr = (int*) malloc(ntarg * sizeof(int));
  *ibox = (int*) malloc(nmax * sizeof(int));
  
  // allocating the locexp array AFTER calculating ntbox!
  
  // centers 
  *center = (double*) malloc(4*nboxes * sizeof(double));
  
  h = 1.0/nside;
  for (i=0;i<nboxes;i++) {
    (*iloc)[i] = 0;
  }
  
  // for each target point...
  int nb2 = nside*nside;
  int nb3 = nb2*nside;
  for (j=0;j<ntarg;j++) {
    // determine which box this point will go into.
    ixh = (int) floor(xtarg[j]/h);
    iyh = (int) floor(ytarg[j]/h);
    izh = (int) floor(ztarg[j]/h);
    iwh = (int) floor(wtarg[j]/h);
    
    if (ixh > nside -1)
      ixh = nside - 1;
    if (ixh < 0)
      ixh = 0;
    if (iyh > nside - 1)
      iyh = nside - 1;
    if (iyh < 0)
      iyh = 0;
    if (izh > nside - 1)
      izh = nside - 1;
    if (izh < 0)
      izh = 0;
    if (iwh > nside - 1)
      iwh = nside - 1;
    if (iwh < 0)
      iwh = 0;
    
    // iadr is the index of the box to which this point is assigned.
    iadr =iwh*nb3 + izh*nb2 + iyh*nside + ixh;
    if ((*iloc)[iadr]==0) {
      (*iloc)[iadr] = (*ntbox);
      // number of target boxes?
      (*ntbox) = (*ntbox) + 1;
    }
  }
  
  
  // allocating the locexp array
  *locexp = (double*****) malloc((*ntbox) * sizeof(double****));
  
  for(i = 0; i < (*ntbox); i++) {
    (*locexp)[i] = (double****) malloc((nterms+1) * sizeof(double***));
    for (j=0; j<(nterms+1); j++) {
      (*locexp)[i][j] = (double***) malloc((nterms+1) * sizeof(double**));
      for (k=0;k<(nterms+1);k++) {
	(*locexp)[i][j][k] = (double**) malloc((nterms+1) * sizeof(double *));
	for (l=0;l<(nterms+1);l++) {
	  (*locexp)[i][j][k][l] = (double*) malloc((nterms+1) * sizeof(double));
	}
      }
    }
  }
  return;
}

 double scale_to_cube(double* sx, double* sy, double* sz,double *sw, int NX){
    int i;
    double maxx = -INF;
    double minx = INF;
    double maxy = -INF;
    double miny = INF;
    double maxz = -INF;
    double minz = INF;
    double maxw = -INF;
    double minw = INF;
    
    double dx, dy, dz, dw;
    double scale;
    for (i = 0; i < NX; i++) {
      maxx = dmax(maxx, sx[i]);
      minx = dmin(minx, sx[i]);
      maxy = dmax(maxy, sy[i]);
      miny = dmin(miny, sy[i]);
      maxz = dmax(maxz, sz[i]);
      minz = dmin(minz, sz[i]);
      maxw = dmax(maxw, sw[i]);
      minw = dmin(minw, sw[i]);
    }
    
    dx = maxx - minx;
    dy = maxy - miny;
    dz = maxz - minz;
    dw = maxw - minw;
    scale = dmax(dx, dy);
    scale = dmax(scale,dz);
    scale = dmax(scale,dw);
    if(scale==0.0)scale=1.0;
    for (i = 0; i < NX; i++) {
        sx[i] = (sx[i] - minx) / scale;
        sy[i] = (sy[i] - miny) / scale;
        sz[i] = (sz[i] - minz) / scale;
	sw[i] = (sw[i] - minw) / scale;
    }
    
    sourcescale = scale;
    return scale;
}



 void allocate(int NX){
   sgx = (double*) malloc(NX * sizeof(double));
   sgy = (double*) malloc(NX * sizeof(double));
   sgz = (double*) malloc(NX * sizeof(double));
   sgw = (double*) malloc(NX * sizeof(double));
  
 }
 
 void gaussth1(double *w, double* kparams,int NX, int D, double* bests, double eps_abs,int IRC) {
   
   double h;
   double scale;
   int i;
   h = dmax(kparams[0],kparams[1]);
   h = dmax(h,kparams[2]);
   h = dmax(h,kparams[3]);
   
      
   for (i = 0; i < NX; i++) {
     sgx[i] *= (h/kparams[0]);
     sgy[i] *= (h/kparams[1]);
     sgz[i] *= (h/kparams[2]);
     sgw[i] *= (h/kparams[3]);
   }
   
   scale = scale_to_cube(sgx, sgy, sgz,sgw, NX);
   gausst(sgx, sgy, sgz,sgw, w, NX, bests, (h*h/(scale*scale)), eps_abs,IRC);
   
 }
 void gaussth2(double *w, double* kparams,int NX, int D, double* bests, double eps_abs,int IRC) {
   
   double h;
   double scale = sourcescale;
   h = dmax(kparams[0],kparams[1]);
   h = dmax(h,kparams[2]);
   h = dmax(h,kparams[3]);
   gausst(sgx, sgy, sgz,sgw, w, NX, bests, (h*h/(scale*scale)), eps_abs,IRC);
   
 }

};
#undef INF
//#undef IRC
#undef K1
#undef NE
#undef PI
